/***********************************************************************
 * draw3d.c
 *
 * Copyright (C) 1997-2002 Klaus Ehrenfried
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * 
 * Klaus Ehrenfried 26-Mar-97
 *
 * 17-May-99 trans10, draw1_line, fill1_triangle:
 *           Sofort Abbruch, wenn ein Punkt ausserhalb des Bildes liegt.
 * 
 * 7-Mar-2000 calc_shade, check_direction, fill1_triangle:
 *            added hidden_color
 * 3-Apr-2000 fill1_triangle:
 *            Nur wenn alle Punkte ausserhalb wird nicht gezeichnet
 ***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <math.h>

#define MAXROWLEN 1024
#define MAXGRIDLINES 64

#include "main.h"
#include "draw3d.h"
#include "lut.h"

#define SWAP(a,b,c) {c=a;a=b;b=c;}
#define NORM(a,b,c) (a*a+b*b+c*c)

/************************************************************************
 * init_lights
 ************************************************************************/

static void init_lights(TASK *a)
{
   double alpha,beta;
   double sin_alpha,cos_alpha;
   double sin_beta,cos_beta;
   int n;

   for (n = 0; n < MAXLIGHTS; n++)
     {
	alpha = (a->light[n].alpha + a->view_alpha) * PI/180.0;
	beta = (a->light[n].beta + a->view_beta) * PI/180.0;

	cos_alpha=cos(alpha);
	sin_alpha=sin(alpha);

	cos_beta=cos(beta);
	sin_beta=sin(beta);

	a->light[n].d[0]=cos_beta*sin_alpha;
	a->light[n].d[1]=-cos_beta*cos_alpha;
	a->light[n].d[2]=sin_beta;

	/*
	 a->light[n].d[0]=cos_beta*cos_alpha;
	 a->light[n].d[1]=cos_beta*sin_alpha;
	 a->light[n].d[2]=sin_beta;
	 */

	if (a->light[n].power == 0.0) {
	   a->light[n].flag=0;
	}
	else {
	   a->light[n].flag=1;
	}
	
	/*printf("Light (%d): %f %f %f  (%f)\n",n,
	       a->light[n].d[0],
	       a->light[n].d[1],
	       a->light[n].d[2],
	       NORM(a->light[n].d[0],a->light[n].d[1],a->light[n].d[2]));*/
     }
}

/************************************************************************
 * calc_shade
 ************************************************************************/

static void calc_shade(double d[3], int c[3], int flag, TASK *a)
{
   int j,n;
   double h,sum;

   sum=0.0;
   for (n=0; n < MAXLIGHTS; n++) {
      if (a->light[n].flag == 1) {
	 h=0.0;
	 for (j=0; j < 3; j++) {
	    h += (d[j] * a->light[n].d[j]);
	 }
	 /*h = 0.5*(1.0+h);*/
	 if (h < 0.0) h = 0.0;
	 sum += (h * a->light[n].power);
      }
   }

   h=1.0 - a->render_ambient;
   if (h > 0) sum=a->render_ambient + sum*h;

   if (flag) {
      for (j=0; j < 3; j++) {
	 h=sum * a->render_color[j];
	 c[j]=256.0 * h - 0.5;
	 if (c[j] < 0) c[j]=0;
	 if (c[j] > 255) c[j]=255;
      }
   } else {
      for (j=0; j < 3; j++) {
	 h=sum * a->render_hidden_color[j];
	 c[j]=256.0 * h - 0.5;
	 if (c[j] < 0) c[j]=0;
	 if (c[j] > 255) c[j]=255;
      }
   }

   /*
   printf("cs: %f %f %f (%f)    %d %d %d\n",
	  d[0],d[1],d[2],sum,
	  c[0],c[1],c[2]);*/
}

/************************************************************************
 * set_line_width
 ************************************************************************/

void set_line_width(int width, TASK *a)
{
  int *p;
  int len;
  int i,j;
  int start,end;

  start=end=width/2;

  if ((start+end) < width) {
    a->line_pixel_shift=1;
    start++;
  }
  else {
    a->line_pixel_shift=0;
  }

  len=0;
  p=a->line_jump;
  a->line_jump_len=0;

  for (i=-start; i <= end; i++)
    {
      for (j=-start; j <= end; j++)
	{
	  if (len >= a->line_jump_max) return;
	  *p++ = i;
	  *p++ = j;

	  len++;
	}
    }

  a->line_jump_len = len;
}

/************************************************************************
 * set0_pixel
 ************************************************************************/

void set0_pixel(int i, int j, double z, int c[3], TASK *a)
{
  PIXEL *pp;

  /*printf("setpixel %d %d\n",i,j);*/
  if ((i < 0) || (i >= a->image_width)) return;
  if ((j < 0) || (j >= a->image_height)) return;

  pp=a->pels + (i + (j * a->image_width));

  if ((pp->hit == 0) || (z < pp->z))
    {
      pp->z = z;
      pp->r = c[0];
      pp->g = c[1];
      pp->b = c[2];
      pp->hit = 1;
    }
}

/************************************************************************
 * draw0_spot
 ************************************************************************/

static void draw0_spot(int i0, int j0, double z, int c[3], TASK *a)
{
   int *p;
   int n;
   int i,j;
   double z1;
   
   z1=z-a->line_dz;
   if (a->line_jump_len <= 0)
     {
	set0_pixel(i0,j0,z1,c,a);
	return;
     }

   p=a->line_jump;
   for (n=0; n < a->line_jump_len; n++)
     {
	i=i0 + *(p++);
	j=j0 + *(p++);
	set0_pixel(i,j,z1,c,a);
     }
}

/************************************************************************
 * draw0_line_const
 ************************************************************************/

void draw0_line_const(RPOINT *p1, RPOINT *p2, int c[3], TASK *a)
{
  int di,dj;
  double adz,adi,adj;
  double base_z,base_i,base_j;
  double steps;
  double z;
  int n;
  int nsteps;
  int i,j;

  di=p2->i - p1->i;
  dj=p2->j - p1->j;
  adz=p2->z - p1->z;

  if ((di == 0) && (dj == 0))
    {
      draw0_spot(p1->i, p1->j, p1->z, c, a);
      draw0_spot(p2->i, p2->j, p2->z, c, a);
    }
  else if (abs(di) > abs(dj))
    {
      steps=di;
      adj=(double)dj/steps;
      adz=adz/steps;
      if (di > 0)
	{
	  base_j=0.5+p1->j;
	  base_z=p1->z;
	  i=p1->i;
	  nsteps=di;
	}
      else
	{
	  base_j=0.5+p2->j;
	  base_z=p2->z;
	  i=p2->i;
	  nsteps=-di;
	}

      for (n=0; n <= nsteps; n++)
	{
	  j=(int)(base_j + n * adj);
	  z=base_z + n * adz;
	  draw0_spot(i, j, z, c, a);
	  i++;
	}
    }
  else
    {
      steps=dj;
      adi=(double)di/steps;
      adz=adz/steps;
      if (dj > 0)
	{
	  base_i=0.5+p1->i;
	  base_z=p1->z;
	  j=p1->j;
	  nsteps=dj;
	}
      else
	{
	  base_i=0.5+p2->i;
	  base_z=p2->z;
	  j=p2->j;
	  nsteps=-dj;
	}

      for (n=0; n <= nsteps; n++)
	{
	  i=(int)(base_i + n * adi);
	  z=base_z + n * adz;
	  draw0_spot(i, j, z, c, a);
	  j++;
	}
    }
}

/************************************************************************
 * draw0_line_clu
 ************************************************************************/

static void draw0_line_clu(RPOINT *p1, RPOINT *p2, TASK *a)
{
   int di,dj;
   double adz,adi,adj;
   double adw;
   double base_z,base_i,base_j;
   double base_w;
   double steps;
   double z;
   double w;
   int c[3];
   int n;
   int nsteps;
   int i,j;

   di=p2->i - p1->i;
   dj=p2->j - p1->j;
   adz=p2->z - p1->z;
   adw=p2->w - p1->w;

   if ((di == 0) && (dj == 0))
     {
	look_up_color(p1->w, c, a);
	draw0_spot(p1->i, p1->j, p1->z, c, a);
	look_up_color(p2->w, c, a);
	draw0_spot(p2->i, p2->j, p2->z, c, a);
     }
   else if (abs(di) > abs(dj))
     {
	steps=di;
	adj=(double)dj/steps;
	adz=adz/steps;
	adw=adw/steps;
	if (di > 0)
	  {
	     base_j=0.5+p1->j;
	     base_z=p1->z;
	     base_w=p1->w;
	     i=p1->i;
	     nsteps=di;
	  }
	else
	  {
	     base_j=0.5+p2->j;
	     base_z=p2->z;
	     base_w=p2->w;
	     i=p2->i;
	     nsteps=-di;
	  }
	
	for (n=0; n <= nsteps; n++)
	  {
	     j=(int)(base_j + n * adj);
	     z=base_z + n * adz;
	     w=base_w + n * adw;
	     look_up_color(w, c, a);
	     draw0_spot(i, j, z, c, a);
	     i++;
	  }
     }
   else
     {
	steps=dj;
	adi=(double)di/steps;
	adz=adz/steps;
	adw=adw/steps;
	if (dj > 0)
	  {
	     base_i=0.5+p1->i;
	     base_z=p1->z;
	     base_w=p1->w;
	     j=p1->j;
	     nsteps=dj;
	  }
	else
	  {
	     base_i=0.5+p2->i;
	     base_z=p2->z;
	     base_w=p2->w;
	     j=p2->j;
	     nsteps=-dj;
	  }

	for (n=0; n <= nsteps; n++)
	  {
	     i=(int)(base_i + n * adi);
	     z=base_z + n * adz;
	     w=base_w + n * adw;
	     look_up_color(w, c, a);
	     draw0_spot(i, j, z, c, a);
	     j++;
	  }
     }
}

/************************************************************************
 * fill0_hor_const
 ************************************************************************/

static void fill0_hor_const(int j0,
			    int i1, double z1,
			    int i2, double z2,
			    int c[3],
			    TASK *a)
{
   double adz;
   double z;
   int n,i;
   int nsteps;

   if (i2 < i1)
     {
	SWAP(i1,i2,i);
	SWAP(z1,z2,z);
     }

   nsteps=i2 - i1;

   if (nsteps == 0)
     {
	set0_pixel(i1, j0, z1, c, a);
	set0_pixel(i2, j0, z2, c, a);
     }
   else
     {
	adz=(z2-z1)/nsteps;
	i=i1;
	for (n=0; n <= nsteps; n++)
	  {
	     z=z1 + n * adz;
	     set0_pixel(i, j0, z, c, a);
	     i++;
	  }
     }
}

/************************************************************************
 * fill0_hor_clu
 ************************************************************************/

static void fill0_hor_clu(int j0,
		   int i1, double z1, double w1,
		   int i2, double z2, double w2,
		   TASK *a)
{
   double adz;
   double adw;
   double z;
   double w;
   int c[3];
   int n,i;
   int nsteps;

   if (i2 < i1)
     {
	SWAP(i1,i2,i);
	SWAP(w1,w2,w);
	SWAP(z1,z2,z);
     }

   nsteps=i2 - i1;

   if (nsteps == 0)
     {
	look_up_color(w1, c, a);
	set0_pixel(i1, j0, z1, c, a);
	look_up_color(w2, c, a);
	set0_pixel(i2, j0, z2, c, a);
     }
   else
     {
	adz=(z2-z1)/nsteps;
	adw=(w2-w1)/nsteps;
	i=i1;
	for (n=0; n <= nsteps; n++)
	  {
	     z=z1 + n * adz;
	     w=w1 + n * adw;
	     look_up_color(w, c, a);
	     set0_pixel(i, j0, z, c, a);
	     i++;
	  }
     }
}

/************************************************************************
 * fill0_trapezoid_const
 ************************************************************************/

static void fill0_trapezoid_const(int j1,
			   int i11, double z11,
			   int i12, double z12,
			   int j2,
			   int i21, double z21,
			   int i22, double z22,
			   int c[3],
			   TASK *a)
{
   int i1,i2,j,n;
   int nsteps;
   double z1,z2;
   double adz1,adz2;
   double adi1,adi2;
   double base_i1,base_i2;
   
   if (j2 < j1)
     {
	SWAP(j1,j2,j);

	SWAP(i11,i21,i1);
	SWAP(i12,i22,i2);

	SWAP(z11,z21,z1);
	SWAP(z12,z22,z2);
     }

   nsteps=j2-j1;
   if (nsteps == 0)
     {
	fill0_hor_const(j1, i11, z11, i12, z12, c, a);
	fill0_hor_const(j2, i21, z21, i22, z22, c, a);
     }
   else
     {
	adz1=(z21-z11)/nsteps;
	adz2=(z22-z12)/nsteps;

	adi1=(double)(i21-i11)/(double)nsteps;
	adi2=(double)(i22-i12)/(double)nsteps;
	base_i1=0.5+i11;
	base_i2=0.5+i12;

     	j=j1;
	for (n=0; n <= nsteps; n++)
	  {
	     z1=z11 + n * adz1;
	     z2=z12 + n * adz2;
	     i1=(int)(base_i1 + n * adi1);
	     i2=(int)(base_i2 + n * adi2);
	     fill0_hor_const(j, i1, z1, i2, z2, c, a);
	     j++;
	  }

     }
}

/************************************************************************
 * fill0_trapezoid_clu
 ************************************************************************/

void fill0_trapezoid_clu(int j1,
			 int i11, double z11, double w11,
			 int i12, double z12, double w12,
			 int j2,
			 int i21, double z21, double w21,
			 int i22, double z22, double w22,
			 TASK *a)
{
   int i1,i2,j,n;
   int nsteps;
   double w1,w2,z1,z2;
   double adz1,adz2;
   double adw1,adw2;
   double adi1,adi2;
   double base_i1,base_i2;
   
   if (j2 < j1)
     {
	SWAP(j1,j2,j);

	SWAP(i11,i21,i1);
	SWAP(i12,i22,i2);

	SWAP(z11,z21,z1);
	SWAP(z12,z22,z2);

	SWAP(w11,w21,w1);
	SWAP(w12,w22,w2);
     }

   nsteps=j2-j1;
   if (nsteps == 0)
     {
	fill0_hor_clu(j1, i11, z11, w11, i12, z12, w12, a);
	fill0_hor_clu(j2, i21, z21, w21, i22, z22, w22, a);
     }
   else
     {
	adz1=(z21-z11)/nsteps;
	adz2=(z22-z12)/nsteps;
	adw1=(w21-w11)/nsteps;
	adw2=(w22-w12)/nsteps;

	adi1=(double)(i21-i11)/(double)nsteps;
	adi2=(double)(i22-i12)/(double)nsteps;
	base_i1=0.5+i11;
	base_i2=0.5+i12;

     	j=j1;
	for (n=0; n <= nsteps; n++)
	  {
	     z1=z11 + n * adz1;
	     z2=z12 + n * adz2;
	     w1=w11 + n * adw1;
	     w2=w12 + n * adw2;
	     i1=(int)(base_i1 + n * adi1);
	     i2=(int)(base_i2 + n * adi2);
	     fill0_hor_clu(j, i1, z1, w1, i2, z2, w2, a);
	     j++;
	  }

     }
}

/************************************************************************
 * fill0_triangle_const
 ************************************************************************/

static void fill0_triangle_const(RPOINT *p1, RPOINT *p2, RPOINT *p3,
			  int c[3], TASK *a)
{
   RPOINT *pmid,*pmax,*pmin;
   double eps,emeps;
   double zeps;
   int ieps;
   int dj0;

   if (p1->j <= p2->j)
     {
	if (p3->j >= p2->j) {pmax=p3;pmid=p2;pmin=p1;}
	else if (p3->j >= p1->j) {pmax=p2;pmid=p3;pmin=p1;}
	else {pmax=p2;pmid=p1;pmin=p3;}
     }
   else
     {
	if (p3->j >= p1->j) {pmax=p3;pmid=p1;pmin=p2;}
	else if (p3->j >= p2->j) {pmax=p1;pmid=p3;pmin=p2;}
	else {pmax=p1;pmid=p2;pmin=p3;}
     }

   dj0=pmax->j - pmin->j;

   if (dj0 == 0)
     {
	eps=0.5;
     }
   else
     {
	eps=(double)(pmid->j - pmin->j)/(double)dj0;
     }

   emeps=1.0 - eps;

   zeps = eps * pmax->z + emeps * pmin->z;
   ieps = eps * pmax->i + emeps * pmin->i + 0.5;

   fill0_trapezoid_const(
			 pmin->j,
			 pmin->i, pmin->z,
			 pmin->i, pmin->z,
			 pmid->j,
			 pmid->i, pmid->z,
			 ieps, zeps,
			 c,a);

   fill0_trapezoid_const(
			 pmid->j,
			 pmid->i, pmid->z,
			 ieps, zeps,
			 pmax->j,
			 pmax->i, pmax->z,
			 pmax->i, pmax->z,
			 c,a);


}

/************************************************************************
 * fill0_triangle_clu
 ************************************************************************/

static void fill0_triangle_clu(RPOINT *p1, RPOINT *p2, RPOINT *p3, TASK *a)
{
   RPOINT *pmid,*pmax,*pmin;
   double eps,emeps;
   double zeps,weps;
   int ieps;
   int dj0;

   if (p1->j <= p2->j)
     {
	if (p3->j >= p2->j) {pmax=p3;pmid=p2;pmin=p1;}
	else if (p3->j >= p1->j) {pmax=p2;pmid=p3;pmin=p1;}
	else {pmax=p2;pmid=p1;pmin=p3;}
     }
   else
     {
	if (p3->j >= p1->j) {pmax=p3;pmid=p1;pmin=p2;}
	else if (p3->j >= p2->j) {pmax=p1;pmid=p3;pmin=p2;}
	else {pmax=p1;pmid=p2;pmin=p3;}
     }

   dj0=pmax->j - pmin->j;

   if (dj0 == 0)
     {
	eps=0.5;
     }
   else
     {
	eps=(double)(pmid->j - pmin->j)/(double)dj0;
     }

   emeps=1.0 - eps;

   zeps = eps * pmax->z + emeps * pmin->z;
   weps = eps * pmax->w + emeps * pmin->w;
   ieps = eps * pmax->i + emeps * pmin->i + 0.5;

   fill0_trapezoid_clu(
		       pmin->j,
		       pmin->i, pmin->z, pmin->w,
		       pmin->i, pmin->z, pmin->w,
		       pmid->j,
		       pmid->i, pmid->z, pmid->w,
		       ieps, zeps, weps,
		       a);

   fill0_trapezoid_clu(
		       pmid->j,
		       pmid->i, pmid->z, pmid->w,
		       ieps, zeps, weps,
		       pmax->j,
		       pmax->i, pmax->z, pmax->w,
		       pmax->i, pmax->z, pmax->w,
		       a);


}

/************************************************************************
 * init_tmatrix
 ************************************************************************/

void init_tmatrix(TASK *a)
{
   double alpha,beta;
   double sin_alpha,cos_alpha;
   double sin_beta,cos_beta;
   double scale;
   int j;

   alpha = a->view_alpha * PI/180.0;
   beta = a->view_beta * PI/180.0;

   cos_alpha=cos(alpha);
   sin_alpha=sin(alpha);

   cos_beta=cos(beta);
   sin_beta=sin(beta);

   a->tmatrix_h[0]=cos_alpha;
   a->tmatrix_h[1]=sin_alpha;
   a->tmatrix_h[2]=0.0;

   a->tmatrix_z[0]=-cos_beta*sin_alpha;
   a->tmatrix_z[1]=cos_beta*cos_alpha;
   a->tmatrix_z[2]=-sin_beta;

   a->tmatrix_v[0]=-sin_beta*sin_alpha;
   a->tmatrix_v[1]=sin_beta*cos_alpha;
   a->tmatrix_v[2]=cos_beta;

   if (a->image_width >= a->image_height) {
      scale=a->image_height;
   }
   else {
      scale=a->image_width;
   }
     
   scale *= (0.5/sqrt(3.0) * a->view_zoom);
   printf("Scale: %f\n",scale);
   
   for (j=0; j < 3; j++) {
      a->tmatrix_h[j] *= scale;
      a->tmatrix_v[j] *= scale;
   }

   /*
   printf("(h) %f %f %f  %f\n",
	  a->tmatrix_h[0],
	  a->tmatrix_h[1],
	  a->tmatrix_h[2],
	  NORM(a->tmatrix_h[0],a->tmatrix_h[1],a->tmatrix_h[2]));

   printf("(v) %f %f %f  %f\n",
	  a->tmatrix_v[0],
	  a->tmatrix_v[1],
	  a->tmatrix_v[2],
	  NORM(a->tmatrix_v[0],a->tmatrix_v[1],a->tmatrix_v[2]));

   printf("(z) %f %f %f  %f\n",
	  a->tmatrix_z[0],
	  a->tmatrix_z[1],
	  a->tmatrix_z[2],
	  NORM(a->tmatrix_z[0],a->tmatrix_z[1],a->tmatrix_z[2]));*/

}

/************************************************************************
 * trans10
 ************************************************************************/

int trans10(POINT *p, RPOINT *rp, int shift, TASK *a)
{
   double h,v,z;
   double d;
   double s;

   z=a->tmatrix_z[0] * p->x +
     a->tmatrix_z[1] * p->y +
     a->tmatrix_z[2] * p->z;

   h=a->tmatrix_h[0] * p->x +
     a->tmatrix_h[1] * p->y +
     a->tmatrix_h[2] * p->z;

   v=a->tmatrix_v[0] * p->x +
     a->tmatrix_v[1] * p->y +
     a->tmatrix_v[2] * p->z;

   /*printf("trans10: %f %f\n",h,v);*/
   d=a->view_dist + z;
   if (d <= 0.0) d=1.0;
   s=a->view_dist/d;

   rp->z = z;
   rp->w = p->w;
   if (shift == 1) {
     rp->i = a->image_center[0] + h * s + 0.5;
     rp->j = a->image_center[1] - v * s + 0.5;
   }
   else {
     rp->i = a->image_center[0] + h * s;
     rp->j = a->image_center[1] - v * s;
   }

   if (rp->i < -a->out_tolerance) return(0);
   if (rp->j < -a->out_tolerance) return(0);
   if (rp->i > a->image_width + a->out_tolerance) return(0);
   if (rp->j > a->image_height + a->out_tolerance) return(0);

   return(1);
}

/************************************************************************
 * draw1_line
 ************************************************************************/

void draw1_line(POINT *p1, POINT *p2, TASK *a)
{
   RPOINT rp1,rp2;

   if (!trans10(p1, &rp1, a->line_pixel_shift, a)) return;
   if (!trans10(p2, &rp2, a->line_pixel_shift, a)) return;

   /*
    rp1.z -= a->line_dz;
    rp2.z -= a->line_dz;
    */

   if (a->draw_clu_flag == 0) {
      draw0_line_const(&rp1, &rp2, a->work_color, a);
   }
   else {
      draw0_line_clu(&rp1, &rp2, a);
   }
}

/************************************************************************
 * calc_norm_vector
 ************************************************************************/

void calc_norm_vector(POINT *p1, POINT *p2, POINT *p3, double v[3])
{
   double dax,day,daz;
   double dbx,dby,dbz;
   double alen;

   dax = p2->x - p1->x;
   day = p2->y - p1->y;
   daz = p2->z - p1->z;
   
   dbx = p3->x - p1->x;
   dby = p3->y - p1->y;
   dbz = p3->z - p1->z;
   
   v[0] = day*dbz-daz*dby;
   v[1] = daz*dbx-dax*dbz;
   v[2] = dax*dby-day*dbx;

   alen = sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
   
   if (alen > 0.0)
     {
	v[0] /= alen;
	v[1] /= alen;
	v[2] /= alen;
     }
}

/************************************************************************
 * check_direction
 ************************************************************************/

static int check_direction(double v[3], TASK *a)
{
   int j;
   double sum;

   sum=0.0;
   for (j=0; j < 3; j++)
     {
	sum += (v[j] * a->tmatrix_z[j]);
     }

   if (sum >= 0.0)
     {
	for (j=0; j < 3; j++)
	  {
	     v[j] = -v[j];
	  }
	return(1);
     }
   return(0);
}

/************************************************************************
 * fill1_triangle
 ************************************************************************/

static void fill1_triangle(POINT *p1, POINT *p2, POINT *p3, TASK *a)
{
   RPOINT rp1,rp2,rp3;
   double v[3];
   int c[3];
   int res;

   //if (!trans10(p1, &rp1, 0, a)) return;
   //if (!trans10(p2, &rp2, 0, a)) return;
   //if (!trans10(p3, &rp3, 0, a)) return;
   res = trans10(p1, &rp1, 0, a)
     + trans10(p2, &rp2, 0, a)
       + trans10(p3, &rp3, 0, a);
   if (res == 0) return;

   if (a->draw_shading_flag == 0) {
      calc_norm_vector(p1, p2, p3, v);
      if (check_direction(v, a)) {
	 fill0_triangle_const(&rp1, &rp2, &rp3, a->work_color, a);
      } else {
	 fill0_triangle_const(&rp1, &rp2, &rp3, a->hidden_color, a);
      }
   }
   else if (a->draw_shading_flag == 1) {
      fill0_triangle_clu(&rp1, &rp2, &rp3, a);
   }
   else {
      calc_norm_vector(p1, p2, p3, v);
      if (check_direction(v, a)) {
	 calc_shade(v, c, 1, a);
      } else {
	 calc_shade(v, c, 0, a);
      }
      fill0_triangle_const(&rp1, &rp2, &rp3, c, a);
   }
}

/************************************************************************
 * draw1_view_box
 ************************************************************************/

void draw1_view_box(TASK *a)
{
   POINT p1,p2;
   double x,y,z;
   int i,j,k;

   /*init_tmatrix(a);*/

   for (j=0; j < 3; j++) {
      a->work_color[j] = a->line_color[j];
   }
   a->draw_clu_flag=0;
   set_line_width(a->line_width,a);

   for (i=0; i < 2; i++) {
      for (j=0; j < 2; j++) {
	 for (k=0; k < 3; k++) {
	    x=a->view_box[0];
	    y=a->view_box[1];
	    z=a->view_box[2];
	    if (k == 0)
	      {
		 p1.x = -x;
		 p2.x = x;

		 if (i == 0) y=-y;
		 if (j == 0) z=-z;

		 p1.y = p2.y = y;
		 p1.z = p2.z = z;
	      }
	    else if (k == 1)
	      {
		 p1.y = -y;
		 p2.y = y;

		 if (i == 0) x=-x;
		 if (j == 0) z=-z;

		 p1.x = p2.x = x;
		 p1.z = p2.z = z;
	      }
	    else
	      {
		 p1.z = -z;
		 p2.z = z;

		 if (i == 0) x=-x;
		 if (j == 0) y=-y;

		 p1.x = p2.x = x;
		 p1.y = p2.y = y;
	      }
	    draw1_line(&p1, &p2, a);
	 }
      }
   }
}

/************************************************************************
 * init_view_box
 ************************************************************************/

void init_view_box(TASK *a)
{
   double delta;
   int j;

   for (j=0; j < 3; j++)
     {
	a->view_origin[j] = 0.5 * (a->view_start[j] + a->view_end[j]);
	delta = a->view_end[j] - a->view_start[j];
	if (delta == 0.0) {
	   a->view_scale[j] = 2.0;
	}
	else {
	   a->view_scale[j] = 2.0*a->view_box[j]/delta;
	}
     }
}

/************************************************************************
 * trans21
 ************************************************************************/

static void trans21(POINT *r, POINT *p, TASK *a)
{
   p->x = (r->x - a->view_origin[0]) * a->view_scale[0];
   p->y = (r->y - a->view_origin[1]) * a->view_scale[1];
   p->z = (r->z - a->view_origin[2]) * a->view_scale[2];
   p->w = r->w;
}

/************************************************************************
 * draw2_line
 ************************************************************************/

static void draw2_line(POINT *r1, POINT *r2, TASK *a)
{
   POINT p1,p2;

   /*printf("Draw: (%f %f %f %f) (%f %f %f %f)\n",
	  r1[0],r1[1],r1[2],r1[3],
	  r2[0],r2[1],r2[2],r2[3]);*/
   trans21(r1, &p1, a);
   trans21(r2, &p2, a);
   draw1_line(&p1, &p2, a);
}

/************************************************************************
 * fill2_triangle
 ************************************************************************/

static void fill2_triangle(POINT *r1, POINT *r2, POINT *r3, TASK *a)
{
   POINT p1,p2,p3;

   trans21(r1, &p1, a);
   trans21(r2, &p2, a);
   trans21(r3, &p3, a);
   fill1_triangle(&p1, &p2, &p3, a);
}

/************************************************************************
 * interpolate_point
 ************************************************************************/

static void interpolate_point(POINT *r1, POINT *r2, POINT *ri, double eps)
{
  double emeps;

  emeps=1.0-eps;
  ri->x = emeps*r1->x + eps*r2->x;
  ri->y = emeps*r1->y + eps*r2->y;
  ri->z = emeps*r1->z + eps*r2->z;
  ri->w = emeps*r1->w + eps*r2->w;
}

/************************************************************************
 * cont2_triangle
 ************************************************************************/

static void cont2_triangle(POINT *r1, POINT *r2, POINT *r3, TASK *a)
{
   POINT ra,rb,rc;
   double wmin,wmax,wlevel;
   double dw21,dw31,dw32;
   double eps;
   int n;
   int mflag;

   wmin=wmax=r1->w;
   if (r2->w < wmin) wmin=r2->w;
   if (r2->w > wmax) wmax=r2->w;

   if (r3->w < wmin) wmin=r3->w;
   if (r3->w > wmax) wmax=r3->w;

   if (wmin == wmax) return;

   dw21=r2->w - r1->w;
   dw31=r3->w - r1->w;
   dw32=r3->w - r2->w;

   for (n=0; n < a->cont_levels; n++) {
      mflag=0;
      wlevel=a->cont_value[n];
      if ((wlevel >= wmin) && (wlevel <= wmax)) {

	 if (dw21 == 0.0) {
	    if (wlevel == r1->w) {
	       draw2_line(r1, r2, a);
	       continue;
	    }
	 }
	 else {
	    eps=(wlevel - r1->w)/dw21;
	    if ((eps >= 0.0) && (eps <= 1.0)) {
	       interpolate_point(r1,r2,&ra,eps);
	       mflag=1;
	    }
	 }

	 if (dw32 == 0.0) {
	    if (wlevel == r2->w) {
	       draw2_line(r2, r3, a);
	       continue;
	    }
	 }
	 else {
	    eps=(wlevel - r2->w)/dw32;
	    if ((eps >= 0.0) && (eps <= 1.0)) {
	       interpolate_point(r2,r3,&rb,eps);
	       if (mflag == 1) {
		  draw2_line(&ra, &rb, a);
	       }
	       mflag=2;
	    }
	 }

	 if (dw31 == 0.0) {
	    if (wlevel == r1->w) {
	       draw2_line(r1, r3, a);
	       continue;
	    }
	 }
	 else {
	    eps=(wlevel - r1->w)/dw31;
	    if ((eps >= 0.0) && (eps <= 1.0)) {
	       interpolate_point(r1,r3,&rc,eps);
	       if (mflag == 1) {
		  draw2_line(&ra, &rc, a);
	       }
	       else if (mflag == 2) {
		  draw2_line(&rb, &rc, a);
	       }
	    }
	 }
      }
   }
}

/************************************************************************
 * cont2_quad
 ************************************************************************/

static void cont2_quad(POINT *r1, POINT *r2, POINT *r3, POINT *r4, TASK *a)
{
  POINT rcenter;

  rcenter.x = 0.25*(r1->x + r2->x + r3->x + r4->x);
  rcenter.y = 0.25*(r1->y + r2->y + r3->y + r4->y);
  rcenter.z = 0.25*(r1->z + r2->z + r3->z + r4->z);
  rcenter.w = 0.25*(r1->w + r2->w + r3->w + r4->w);

  cont2_triangle(r1,r2,&rcenter,a);
  cont2_triangle(r2,r3,&rcenter,a);
  cont2_triangle(r3,r4,&rcenter,a);
  cont2_triangle(r4,r1,&rcenter,a);
}

/************************************************************************
 * fill1_quad
 ************************************************************************/

static void fill1_quad(POINT *r1, POINT *r2, POINT *r3, POINT *r4, TASK *a)
{
  POINT rcenter;

  rcenter.x = 0.25*(r1->x + r2->x + r3->x + r4->x);
  rcenter.y = 0.25*(r1->y + r2->y + r3->y + r4->y);
  rcenter.z = 0.25*(r1->z + r2->z + r3->z + r4->z);
  rcenter.w = 0.25*(r1->w + r2->w + r3->w + r4->w);

  fill1_triangle(r1,r2,&rcenter,a);
  fill1_triangle(r2,r3,&rcenter,a);
  fill1_triangle(r3,r4,&rcenter,a);
  fill1_triangle(r4,r1,&rcenter,a);
}

/************************************************************************
 * draw1_grid
 ************************************************************************/

void draw1_grid(double r, int n1, int n2, int dir, TASK *a)
{
   POINT pa1,pa2,pb1,pb2;
   double h1[MAXGRIDLINES];
   double h2[MAXGRIDLINES];
   double deps;
   int istart,iend,istep;
   int jstart,jend,jstep;
   int i,j,k1,k2,m;

   if (n1 <= 0) n1=1;
   if (n2 <= 0) n2=1;

   if (n1 >= MAXGRIDLINES) n1 = MAXGRIDLINES-1;
   if (n2 >= MAXGRIDLINES) n2 = MAXGRIDLINES-1;

   pa1.w = pa2.w = pb1.w = pb2.w = 0.0;

   if (dir == 1)
     {
       pa1.x = pa2.x = pb1.x = pb2.x = r * a->view_box[0];
       k1=1;
       k2=2;
     }
   else if (dir == 2)
     {
       pa1.y = pa2.y = pb1.y = pb2.y = r * a->view_box[1];
       k1=0;
       k2=2;
     }
   else
     {
       pa1.z = pa2.z = pb1.z = pb2.z = r * a->view_box[2];
       k1=0;
       k2=1;
     }

   deps=2.0/(double)n1;
   for (i=0; i <= n1; i++)
     {h1[i]=(-1.0+i*deps)*a->view_box[k1];}

   deps=2.0/(double)n2;
   for (i=0; i <= n2; i++)
     {h2[i]=(-1.0+i*deps)*a->view_box[k2];}


   if (a->grid_faces == 1) {
     for (j=0; j < 3; j++) {
       a->work_color[j] = a->face_color[j];
     }
     a->draw_shading_flag=0;

     for (i=0; i < n1; i++)
       {
	 if (dir == 1)
	   {
	     pa1.y = pa2.y = h1[i];
	     pb1.y = pb2.y = h1[i+1];
	   }
	 else
	   {
	     pa1.x = pa2.x = h1[i];
	     pb1.x = pb2.x = h1[i+1];
	   }
	 for (j=0; j < n2; j++)
	   {
	     if (dir == 3)
	       {
		 pa1.y = pb1.y = h2[j];
		 pa2.y = pb2.y = h2[j+1];
	       }
	     else
	       {
		 pa1.z = pb1.z = h2[j];
		 pa2.z = pb2.z = h2[j+1];
	       }

	     fill1_quad(&pa1,&pa2,&pb2,&pb1,a);
	   }
       }
   }

   for (j=0; j < 3; j++) {
     a->work_color[j] = a->line_color[j];
   }
   set_line_width(a->line_width,a);
   a->draw_clu_flag=0;

   for (m=0; m < 2; m++)
     {
       if (m == 0)
	 {
	   istart=1;
	   iend=n1-1;
	   istep=1;

	   jstart=1;
	   jend=n2-1;
	   jstep=1;
	 }
       else
	 {
	   istart=0;
	   iend=n1;
	   istep=n1;

	   jstart=0;
	   jend=n2;
	   jstep=n2;

	   if (a->grid_border == 1) {
	     for (j=0; j < 3; j++) {
	       a->work_color[j] = a->border_color[j];
	     }
	     set_line_width(a->border_width,a);
	   }
	 }

       for (i=istart; i <= iend; i += istep)
	 {
	   if (dir == 1)
	     {
	       pa1.y = pa2.y = h1[i];
	     }
	   else
	     {
	       pa1.x = pa2.x = h1[i];
	     }
	   for (j=0; j < n2; j++)
	     {
	       if (dir == 3)
		 {
		   pa1.y = h2[j];
		   pa2.y = h2[j+1];
		 }
	       else
		 {
		   pa1.z = h2[j];
		   pa2.z = h2[j+1];
		 }
	       draw1_line(&pa1, &pa2, a);
	     }
	 }

       for (i=0; i < n1; i++)
	 {
	   if (dir == 1)
	     {
	       pa1.y = h1[i];
	       pb1.y = h1[i+1];
	     }
	   else
	     {
	       pa1.x = h1[i];
	       pb1.x = h1[i+1];
	     }
	   for (j=jstart; j <= jend; j += jstep)
	     {
	       if (dir == 3)
		 {
		   pa1.y = pb1.y = h2[j];
		 }
	       else
		 {
		   pa1.z = pb1.z = h2[j];
		 }
	       draw1_line(&pa1, &pb1, a);
	     }
	 }
     }

}

/************************************************************************
 * fill2_quad
 ************************************************************************/

static void fill2_quad(POINT *r1, POINT *r2, POINT *r3, POINT *r4, TASK *a)
{
  POINT p1,p2,p3,p4;

  trans21(r1, &p1, a);
  trans21(r2, &p2, a);
  trans21(r3, &p3, a);
  trans21(r4, &p4, a);

  fill1_quad(&p1,&p2,&p3,&p4,a);
}

/************************************************************************
 * get_range
 ************************************************************************/

static void get_range(POINT *p, int len, double rmin[3], double rmax[3])
{
   int i,j;

   if (len <= 0) {
      for (j=0; j < 3; j++) {
	 rmin[j]=-1.0;
	 rmax[j]=1.0;
      }
      return;
   }

   rmin[0]=rmax[0]=p->x;
   rmin[1]=rmax[1]=p->y;
   rmin[2]=rmax[2]=p->z;
   p++;

   for (i=1; i < len; i++) {
     if (rmin[0] > p->x) rmin[0]=p->x;
     if (rmin[1] > p->y) rmin[1]=p->y;
     if (rmin[2] > p->z) rmin[2]=p->z;

     if (rmax[0] < p->x) rmax[0]=p->x;
     if (rmax[1] < p->y) rmax[1]=p->y;
     if (rmax[2] < p->z) rmax[2]=p->z;
     p++;
   }
}

/************************************************************************
 * draw2_face_object
 ************************************************************************/

void draw2_face_object(FACE_OBJECT *obj, TASK *a)
{
   POINT *p1,*p2,*p3,*p4;
   double rmin[3];
   double rmax[3];
   FACE *pf;
   int i,j;

   /*init_tmatrix(a);*/

   get_range(obj->pv, obj->v_count, rmin, rmax);

   for (j=0; j < 3; j++) {
      if (a->view_auto_scale[j] == 1) {
	 a->view_start[j] = rmin[j];
	 a->view_end[j] = rmax[j];
      }
   }

   init_view_box(a);

   if (a->object_faces == 1) {
      if (a->face_shading == 2) {
	for (j=0; j < 3; j++) {
	   a->render_color[j]=(double)a->face_color[j]/255.0;
	   a->render_hidden_color[j]=(double)a->hidden_color[j]/255.0;
	}
	init_lights(a);
      }

      for (j=0; j < 3; j++) {
	 a->work_color[j] = a->face_color[j];
      }
      a->draw_shading_flag = a->face_shading;
      printf("Faces....\n");
      
      pf=obj->pf;
      for (i=0; i < obj->f_count; i++)
	{
	   p1 = obj->pv + pf->id[0];
	   p2 = obj->pv + pf->id[1];
	   p3 = obj->pv + pf->id[2];
	   p4 = obj->pv + pf->id[3];

	   if (pf->id[2] == pf->id[3]) {
	     fill2_triangle(p1,p2,p3,a);
	   }
	   else {
	     fill2_quad(p1,p2,p3,p4,a);
	   }
	   pf++;
	}
   }

   if (a->object_edges == 1) {
      for (j=0; j < 3; j++) {
	 a->work_color[j] = a->edge_color[j];
      }
      a->draw_clu_flag = a->edge_shading;
      set_line_width(a->edge_width,a);
      printf("Edges....\n");
      
      pf=obj->pf;
      for (i=0; i < obj->f_count; i++)
	{
	   p1 = obj->pv + pf->id[0];
	   p2 = obj->pv + pf->id[1];
	   p3 = obj->pv + pf->id[2];
	   p4 = obj->pv + pf->id[3];

	   draw2_line(p1, p2, a);
	   draw2_line(p2, p3, a);
	   draw2_line(p3, p4, a);
	   draw2_line(p4, p1, a);

	   pf++;
	}
   }

   if (a->object_cont == 1) {
      for (j=0; j < 3; j++) {
	 a->work_color[j] = a->line_color[j];
      }
      a->draw_clu_flag = a->line_shading;
      set_line_width(a->line_width,a);
      printf("Contours....\n");
      pf=obj->pf;
      for (i=0; i < obj->f_count; i++)
	{
	   p1 = obj->pv + pf->id[0];
	   p2 = obj->pv + pf->id[1];
	   p3 = obj->pv + pf->id[2];
	   p4 = obj->pv + pf->id[3];

	   if (pf->id[2] == pf->id[3]) {
	      cont2_triangle(p1,p2,p3,a);
	   }
	   else {
	      cont2_quad(p1,p2,p3,p4,a);
	   }
	   pf++;
	}
   }
}

/************************************************************************
 * draw2_mesh_object
 ************************************************************************/

void draw2_mesh_object(MESH_OBJECT *mesh, TASK *a)
{
   POINT *p1,*p2,*p3,*p4;
   double rmin[3];
   double rmax[3];
   int *id;
   int i,j;

   /*init_tmatrix(a);*/

   get_range(mesh->pv, mesh->len, rmin, rmax);

   for (j=0; j < 3; j++) {
      if (a->view_auto_scale[j] == 1) {
	 a->view_start[j] = rmin[j];
	 a->view_end[j] = rmax[j];
      }
   }

   init_view_box(a);

   if ((a->object_faces == 1) && (mesh->rect_flag == 1)){
      if (a->face_shading == 2) {
	for (j=0; j < 3; j++) {
	   a->render_color[j]=(double)a->face_color[j]/255.0;
	   a->render_hidden_color[j]=(double)a->hidden_color[j]/255.0;
	}
	init_lights(a);
      }

      for (j=0; j < 3; j++) {
	 a->work_color[j] = a->face_color[j];
      }
      a->draw_shading_flag = a->face_shading;
      printf("Faces....\n");
      
      id=mesh->id;
      for (j=0; j < (mesh->m-1); j++)
	{
	  p1=mesh->pv + id[j];
	  p2=mesh->pv + id[j+1];
	  p3=p2+1;
	  p4=p1+1;

	  for (i=0; i < (mesh->n-1); i++)
	    {
	      fill2_quad(p1,p2,p3,p4,a);
	      p1=p4;
	      p2=p3;
	      p3++;
	      p4++;
	    }
	}
   }

   if (a->object_edges == 1) {
      for (j=0; j < 3; j++) {
	 a->work_color[j] = a->edge_color[j];
      }
      a->draw_clu_flag = a->edge_shading;
      set_line_width(a->edge_width,a);
      printf("Edges....\n");
      
      id=mesh->id;
      for (j=0; j < mesh->m; j++)
	{
	  p1=mesh->pv + id[j];
	  p2=p1+1;

	  for (i=id[j]; i < (id[j+1]-1); i++)
	    {
	      draw2_line(p1, p2, a);
	      p1=p2;
	      p2++;
	    }
	}
      if (mesh->rect_flag == 1) {
	for (j=0; j < (mesh->m-1); j++)
	  {
	    p1=mesh->pv + id[j];
	    p2=mesh->pv + id[j+1];

	    for (i=0; i < mesh->n; i++)
	      {
		draw2_line(p1, p2, a);
		p1++;
		p2++;
	      }
	  }
      }
   }

   if ((a->object_cont == 1) && (mesh->rect_flag == 1)){
      for (j=0; j < 3; j++) {
	 a->work_color[j] = a->line_color[j];
      }
      a->draw_clu_flag = a->line_shading;
      set_line_width(a->line_width,a);
      printf("Contours....\n");
      
      id=mesh->id;
      for (j=0; j < (mesh->m-1); j++)
	{
	  p1=mesh->pv + id[j];
	  p2=mesh->pv + id[j+1];
	  p3=p2+1;
	  p4=p1+1;

	  for (i=0; i < (mesh->n-1); i++)
	    {
	      cont2_quad(p1,p2,p3,p4,a);
	      p1=p4;
	      p2=p3;
	      p3++;
	      p4++;
	    }
	}
   }

}
